/*
 * Copyright(c) 1997,1998 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 */

#ifndef	_format_h_
#define	_format_h_

extern	char	*default_format_str[];
extern	int	out_sprintf(char *, char *, char *, char *, char *, char *);
extern	char	*sic_split(char *, char *);

#define	F_NUM		0
#define F_SNOTICE	1
#define F_SMSG		2

#define	F_VNOTICE	3

#define F_UNOTICE	4
#define F_UMSG		5
#define F_NICK		6
#define	F_UMODE		7

#define F_CNOTICE	8
#define F_CMSG		9
#define	F_CMODE		10
#define	F_SMODE		11

#define F_INVITE	12
#define F_JOIN		13
#define F_PART		14
#define	F_QUIT		15
#define	F_KICK		16
#define	F_TOPIC		17

#define	F_DCC		18

#define	F_KILL		19
#define	F_ERROR		20
#define	F_UNKNOWN	21

#define	F_MYPRIVM	22
#define	F_MYPRIVN	23
#define	F_MYPUBM	24
#define	F_MYPUBN	25
#define	F_MYQUERY	26
#define	F_MYDCC		27

/* the following 2 aren't formatted with the "rich" routine */
#define	F_SPLIT		28	/* needs 2 %s */
#define	F_PREFIX	29	/* time stuff */

#define	F_MAX		29

#define	N_MAX		599

#endif
