/*
 * Copyright(c) 1997,1998 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 *
 * $Id: acconfig.h,v 1.6 1998/06/04 17:56:10 kalt Exp $
 *
 */

/* Define if sys_errlist[] is declared in stdio.h or errno.h. */
#undef SYS_ERRLIST_DECLARED

/* Define if sys_nerr is declared in stdio.h or errno.h. */
#undef SYS_NERR_DECLARED

/* How do we get a 32bit integer? */
#undef UNSIGNED_LONG32
#undef UNSIGNED_INT32
#undef UNSIGNED_UNKNOWN32

/* Define if compiling on SunOS 5.x */
#undef SUNOS_5

/* Define if using GNU Rx library */
#undef USE_GNU_RX

/* Define if regular expression matching functions are there */
#undef HAVE_REGEXP

