/*
 * Copyright(c) 1997,1998 - All Rights Reserved
 *
 * See the COPYRIGHT file.
 */

void	dns_lookup(char *);

struct server_	*select_server(char *);
struct server_	*sic_server(char *, char *, int, char *);
int	sic_select();
void	vsic_write(char *, ...);

int	cmd_server(char *);

void	vsic_dwrite(char *, char *, ...);
void	sic_dcc(int, char *, unsigned long,unsigned int,char *,unsigned long);
int	cmd_dcc(char *);
